_ENV = AIContext

--locals we have to duplicate...
local function distance_rank(star)

  --if empire.stars[star] 
  --  return 0
  --end

  local ship = get_max_key(free_combat_ships,|) star_distance_value(star)

  if not ship
    return
  end

  local min_dist = fleet_travel_dist(ship2fleet[ship], star)

  return min_dist

end

local function calc_min_attack_power(info)
  local threat, min_defenses in info
  return  attack_caution_mult * max(small_power, (min_defenses+threat), minimum_strike_power)
end

local function value_attack(k,info)
  local star,value,  requires_er in info

  --print(star.name,star.value)

  -- don't attack useless systems
  if (value or 0) <= 0
      return
  end

  local min_power = calc_min_attack_power(info)


  local uncommitted = max_additional_power(star) - reserved_defense_power()

  --[[
  print(star.name, ..min_power, ..threat, .. uncommitted)
  print(..current_defense_power,..minimum_defensive_power)
  print('max additional',max_additional_power(star))
  --]]

  if (uncommitted == 0) and (committed_power[star] or 0 ==0)
    --print('no additional power',star.name)
    return
  end
  

  if not requires_er
    local eta = distance_rank(star!)

    if not eta
      return
    end

    if eta==0
      value += 100
    elseif eta==1
      value += 50
    elseif eta==2
      value += 20
    elseif eta > 3
      value -= eta*10
    end


  end


  if uncommitted >= min_power
    return value * (requires_er and .5 or 1)
  end

  --print('INSUFFICIENT POWER!',star.name, ..min_power, ..uncommitted )

end



function choose_ships.attack()

  if purely_defensive_posture()
    return
  end

  local star, value = get_max_key(attack_priorities,value_attack)

  if not star
    return
  end

  if not next | free_ships
    return
  end

  return value, function()
    
    local info = attack_priorities![star]
    local min_power = calc_min_attack_power(info)
    local danger = min_power - (committed_power[star] or 0)    

    attack_priorities[star] = nil

    local using_ships = {}

    claim_commited_ships(star, danger, using_ships)
    claim_combat_ships_at_star(star, danger, using_ships)

    --print(..danger)

    local total_power = total_ship_power(using_ships)

    local delay_years

    for i,ship in ipairs | FLAG_call('avoid_in_transit',closest_combat_ships,star)

      if total_power >= min_power
        break
      end
      
      -- closest_combat_ships should sort any damaged ships last, however if we
      -- absolutely must include damaged ships, check if any of them are at a place
      -- where they can be repaired next turn. If so, delay the attack by one turn
      if (delay_years or 0) < 1 and damaged_ship(ship) > 0.2
        local src_star = at_star(ship2fleet[ship])
        if src_star and src_star ~= star and src_star:empire_has_repair(empire)
          delay_years = 1
        end
      end

      -- claim it
      using_ships[ship]=true
      total_power += ship_power_score(ship)

      ::next_ship::

    end

    --add_closest_combat_ships(star, max_power, using_ships)

    --print(..target_power,..t2s | using_ships)

    --print | t2s| info


    local _TRACE_INFO = star.name


    --if total_power < min_power
    --  errorf ('thats odd %d %d %d', total_power, min_power, max_additional_power(star))
    --end



    if not next| using_ships
      ai_debug_printf('not adding attack ships for %s',star.name)  
      ai_debug_print('looks like the ships traveling to this star already have another plan')
      return      
    end


    local plan = { 
      mission='attack', 
      ..using_ships, 
      ..star, 
      ..value, 
      ..min_power ,
      sending_power = total_power!,
      ..minimum_strike_power,
      ..attack_caution_mult,
      ..delay_years,
    }

    

    ai_debug_print(t2s | plan)

    for ship ; using_ships
      mark_ship_dest(ship, star, plan)
    end

--    calc_arrival_year(plan)

    -- if we have at least 1 valid attack target, we want to try
    -- it.  however, after that point, we only want additional
    -- attacks if doing so does not overly fragment our fleet.
    --
    update_minimum_strike_power()

    return plan
  end

end