_ENV = AIContext

local function combat_ship_ordering(star)

  return function(s1,s2)

    local d1 = fleet_travel_dist(ship2fleet[s1], star)
    local d2 = fleet_travel_dist(ship2fleet[s2], star)

    local h1,h2=0,0
    if not FLAG.ignore_damage
      -- bin ship damage into discrete steps so that distance still plays a role
      local damage_bucket(d)
        if d < 0.2 return 0 end
        if d < 0.5 return 1 end
        if d < 0.7 return 2 end
        return 3
      end

      h1 = damage_bucket | damaged_ship | s1
      h2 = damage_bucket | damaged_ship | s2
    end

    local f1,f2=0,0

    if FLAG.avoid_in_transit
      f1 = at_star(ship2fleet[s1]) and 0 or 1
      f2 = at_star(ship2fleet[s2]) and 0 or 1
    end

    return lexical_compare(
      f1,f2,
      h1,h2,
      s1.scanner and 1 or 0, s2.scanner and 1 or 0, -- prefer non-scouts
      d1,d2, -- smaller distances
      ship_power_score | s2, ship_power_score | s1, -- larger power
      s1._ikey, s2._ikey) -- unambiguous ordering (favor newer ships)

  end

end

function commited_combat_ships(star)
  local push, ships = pusher {}

  for ship ; fixed_combat_ships
    if ship2fleet[ship].dest_star==star
      push | ship
    end
  end

  table.sort(ships, combat_ship_ordering(star) )

  return ships
end

function closest_combat_ships(star)
  if not empire:can_reach(star)
    return {}
  end

  local push, ships = pusher {}

  local requires_er = dest_requires_er(star)

  for ship ; free_combat_ships
    if (not requires_er) or ship.extended_range
      push | ship
    end
  end

  table.sort(ships, combat_ship_ordering(star) )

  return ships
end


local safe_civilian_plans = set('guard','fallback') --fixed typo
local time_sensitive_plans = set( 'attack', 'colony','scout' )

local high_priority_plans = set('attack','colony')

function route_stragglers()
  route_assault_transports()
  local _TRACE_INFO = empire.name

  for ship ; free_ships


    --ai_debug_print('looking for plan for',ship)

    for plan in rpairs( all_plans, |) function(p0,p1)
        
        return lexical_compare(
          high_priority_plans[p0.mission] and 0 or 1, 
            high_priority_plans[p1.mission] and 0 or 1, 


          fleet_travel_dist(ship2fleet[ship], p0!.star),
           ( fleet_travel_dist(ship2fleet[ship], p1!.star) )
        )
         
      end


      local star in plan
      print(empire.name, ship.name, damaged_ship(ship))
      if noncombat_ship(ship) or damaged_ship(ship) > 0.5
      
        if not safe_civilian_plans[plan.mission]
          goto next_plan
        else
          mark_ship_dest(ship, star, plan)
          goto next_ship
        end

      elseif plan.mission == 'fallback'
        -- military ships want to be involved in attacks or defenses.
        goto next_plan
      end

      local ship_arrival=ship_arrival_year(ship,star)

      if star and (ship.extended_range or not dest_requires_er(star) )

        if time_sensitive_plans[plan.mission]
          if plan.arrival_year>=ship_arrival
            --print('assigning',ship.name,plan.mission,star.name)
            mark_ship_dest(ship, star, plan)
            goto next_ship
          end
        else

          --print('assigning',ship.name,plan.mission,star.name)
          mark_ship_dest(ship, star, plan)
          goto next_ship
        end


      end

      ::next_plan::

    end

    local plan = fallback_plan(ship)
    if plan
       --print('creating a plan for',ship.name)
       mark_ship_dest(ship, plan.star, plan)
    end

    ::next_ship::

  end

end


function calc_arrival_year(plan)

  local star, using_ships in plan!

  local year

  for ship ; using_ships
    year = generic_max( year, ship_arrival_year(ship,star) )
  end

  year += plan.delay_years or 0

  plan.arrival_year = year
end
